from dataclasses import dataclass
from functools import partial
from unittest.mock import patch

from django.contrib.auth import get_user_model, get_user
from django.contrib.messages import get_messages
from django.test import TestCase
from django.urls import reverse

from workos_login.models import LoginRule, LoginMethods, UserLogin
import workos

from workos_login.utils import jit_create_user

workos.api_key = "mock"
workos.client_id = "mock"


@dataclass
class MockProfile:
    profile: dict
    def to_dict(self):
        return self.profile

def get_profile_and_token(code, **profile_attrs):
    """Assume code is the email"""
    return MockProfile({
        "profile": {
           "id": code,
           "email": code,
           "first_name": code,
           "last_name": code,
           "idp_id": code,
           "raw_attributes": {},
            "organization_id": None,
            "connection_id": None,
            **profile_attrs
        }
    })

def get_and_clear_messages(resp, client):
    messages = [x.message for x in get_messages(resp.wsgi_request)]
    try:
        del client.cookies['messages']
    except KeyError:
        pass
    return messages

#sso_mock = patch('workos.client.sso.get_profile_and_token', wraps=get_profile_and_token)
#sso_mock.side_effect = get_profile_and_token

class LoginRuleTest(TestCase):
    def setUp(self) -> None:
        pass

    def test_priority(self):
        get_user_model().objects.create_user(
            first_name="Test",
            last_name="MFA",
            email="test@mfa.com",
            username="test"
        )
        priority_one = LoginRule.objects.create(
            name="Pri 1",
            method=LoginMethods.SAML_SSO,
            connection_id="1",
            priority=1,
            email_regex="test@sso.com"
        )

        priority_two = LoginRule.objects.create(
            name="Pri 2",
            method=LoginMethods.MFA,
            priority=2,
            email_regex="@mfa.com"
        )

        priority_three = LoginRule.objects.create(
            name="Pri 3",
            method=LoginMethods.USERNAME,
            priority=3,
            email_regex="@mfa.com"
        )

        rule = LoginRule.objects.find_rule_for_username("test")
        self.assertEqual(rule, priority_two)

    @patch('workos.client.sso.get_profile_and_token', wraps=partial(get_profile_and_token, organization_id="sso_org", id="1234"), org_id="org_id")
    def test_idp_lookup(self, mock_get_profile):
        """
        Test that idp will find a user correctly using SSO callback
        """
        user = get_user_model().objects.create_user(
            username="stale_username",
            email="test@example.com"
        )
        resp = self.client.get(reverse('sso_callback'), data={"code": "test@example.com"})
        message = get_and_clear_messages(resp, self.client)[0]
        self.assertEqual(message, "Unknown SSO connection or organization")
        self.assertFalse(get_user(self.client).is_authenticated)

        # Add a user login with SSO
        LoginRule.objects.create(
            name="Saml sso",
            method=LoginMethods.SAML_SSO,
            priority=2,
            organization_id="sso_org",
            email_regex="@example.com"
        )
        self.assertFalse(UserLogin.objects.all().exists())
        resp = self.client.get(reverse('sso_callback'), data={"code": "test@example.com"})
        self.assertEqual(get_and_clear_messages(resp, self.client), [])
        self.assertTrue(UserLogin.objects.all().exists())
        self.assertEqual(get_user(self.client), user)
        self.client.logout()
        self.assertFalse(get_user(self.client).is_authenticated)

        # Now try changing the email and ensure the login still works since the ID has not changed
        resp = self.client.get(reverse('sso_callback'), data={"code": "invalidemail@invalid.com"})
        self.assertEqual(get_user(self.client), user)
        self.assertTrue(get_user(self.client).is_authenticated)

    def test_jit_creation(self):
        sso_rule = LoginRule.objects.create(
            name="JIT Creation",
            method=LoginMethods.SAML_SSO,
            jit_creation=True,
            priority=3,
            saved_attributes={
                "username": "{{profile.first_name}}{{profile.last_name}}"
            }
        )
        profile = {
            "first_name": "Santana",
            "last_name": "Clause",
            "email": "sclause@northpole.net",
            "id": "123",
            "idp_id": "333"
        }
        user = jit_create_user(sso_rule, profile)
        self.assertEqual(user.is_active, True)
        self.assertEqual(user.username, "SantanaClause")



# Generated by Django 4.0.6 on 2022-08-15 00:43

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import workos_login.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserLogin',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('mfa_factor', models.CharField(blank=True, max_length=255)),
                ('mfa_type', models.CharField(blank=True, choices=[('sms', 'SMS'), ('totp', 'Authenticator App')], max_length=10)),
                ('sso_id', models.CharField(blank=True, max_length=255)),
                ('idp_id', models.CharField(blank=True, max_length=255)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='LoginRule',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Name for this config', max_length=255, unique=True)),
                ('lookup_attributes', models.JSONField(blank=True, null=True)),
                ('totp_organization_name', models.CharField(blank=True, help_text='The name of organization which shows in authenticator apps.', max_length=255)),
                ('method', models.CharField(choices=[('mfa', 'MFA'), ('magic', 'Passwordless Email'), ('GoogleOAuth', 'Google SSO'), ('MicrosoftOAuth', 'Microsoft SSO'), ('sso', 'SAML SSO'), ('username', 'Username/Password')], max_length=15)),
                ('jit_creation', models.BooleanField(default=False, help_text='If enabled a user account will be automatically created if one does not exist.', verbose_name='Just in time account creation')),
                ('priority', models.IntegerField(default=100, help_text='Priority of this rule. Lower numbers are checked first', unique=True)),
                ('email_regex', workos_login.models.EmailRegexField(blank=True, help_text='Regex to test on email. Ex. ^.+@(domain|domain2)\\.com$', verbose_name='Email Regular Expression Pattern')),
                ('connection_id', models.CharField(blank=True, help_text='WorkOS connection ID', max_length=255)),
                ('organization_id', models.CharField(blank=True, help_text='WorkOS organization ID', max_length=255)),
                ('jit_attributes', models.JSONField(blank=True, default=dict, help_text='Attributes to set on user instance when creating user')),
                ('jit_groups', models.ManyToManyField(blank=True, help_text='Groups a user should be added to if user is being created.', to='auth.group')),
            ],
            options={
                'ordering': ['priority'],
            },
        ),
    ]



import json
from typing import Optional, Any

from django.contrib.auth import get_user_model
from django.core import signing
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned, FieldDoesNotExist
from django.template import Template, Context
from workos.exceptions import BadRequestException

from workos_login.conf import conf
from django.db import models, transaction
import workos

from workos_login.exceptions import RelationDoesNotExist


def get_users():
    return get_user_model().objects.filter(**conf.WORKOS_ACTIVE_USER_FILTER)


def find_user_by_email(email: str) -> Optional[models.Model]:
    """
    Given an email address find a user account that matches (case insensitive).
    If there are duplicates, none are returned.

    :param email: Email address
    :return: A user object or None (if duplicates or none found)
    """
    try:
        return get_users().get(email__iexact=email)
    except (ObjectDoesNotExist, MultipleObjectsReturned):
        return None


def find_user(username_or_email: str) -> Optional[models.Model]:
    """
    Find a user given the username/email they entered in the form.
    If username is an exact match return that.
    If username is not found, and email lookup is enabled - look for the email (case insensitive).
    :param username_or_email: The username entered in the login form
    :return: User if found else None
    """
    user = None
    if conf.WORKOS_USERNAME_LOOKUP:
        try:
            user = get_users().get(username=username_or_email)
        except ObjectDoesNotExist:
            pass
    if not user and conf.WORKOS_EMAIL_LOOKUP:
        user = find_user_by_email(username_or_email)
    return user


def user_has_mfa_enabled(user: models.Model) -> bool:
    from workos_login.models import UserLogin
    try:
        user_login = UserLogin.objects.get(user=user)
        return user_login.mfa_enabled and user_login.mfa_factor
    except UserLogin.DoesNotExist:
        return False


def get_user_login_model(user: models.Model) -> models.Model:
    """
    Get or create a user login model
    """
    from workos_login.models import UserLogin
    try:
        return UserLogin.objects.get(user=user)
    except UserLogin.DoesNotExist:
        return UserLogin.objects.create(user=user)


def has_user_login_model(user: models.Model) -> bool:
    """
    Check to see if a user login model exists.
    :param user: The user object
    :return: True if user model exists, false otherwise.
    """
    from workos_login.models import UserLogin
    return UserLogin.objects.filter(user=user).exists()


def totp_verify_code(factor_id: str, code: str) -> bool:
    challenge_id = workos.client.mfa.challenge_factor(authentication_factor_id=factor_id).get("id")
    return verify_challenge(challenge_id, code)


def verify_challenge(challenge_id: str, code: str) -> bool:
    """
    Given a challenge ID verify that it is correct
    :param challenge_id: challenge id to verify
    :param code: user entered code
    :return: True if valid, False otherwise
    """
    try:
        response = workos.client.mfa.verify_challenge(authentication_challenge_id=challenge_id, code=code)
    except BadRequestException as e:
        return False

    if not response.get("valid", False):
        return False
    return True


def mfa_enroll(user: models.Model, factor_id: str, mfa_type: str) -> None:
    """
    Call this once MFA has been successfully enrolled and at least one verification has passed
    :param user: user to enroll
    :param mfa_type: type of MFA ("sms" or "totp")
    :param factor_id: factor_id to use
    """
    user_login = get_user_login_model(user)
    user_login.mfa_type = mfa_type
    user_login.mfa_factor = factor_id
    user_login.save()


def render_attribute(value: Any, profile: dict) -> Any:
    """
    Helper to render an attribute that will convert strings to templates that are provided with `profile`.
    If value is a dictionary it will nest down and render the leaf values.
    :param value: the attribute value to render
    :param profile: the profile dict passed from WorkOS
    :return: Rendered value to save
    """
    if isinstance(value, dict):
        for k, v in value.items():
            value[k] = render_attribute(v, profile)
    if isinstance(value, str):
        t = Template(value)
        c = Context({"profile": profile})
        return t.render(c)
    return value


def _link_attributes(obj: models.Model, attributes: Optional[dict], profile: dict) -> None:
    """
    Given a user object link any attributes that are needed.

    :param obj: A user object to link related objects to
    :param attributes: The saved attributes that will be used to try and link attributes.
    :param profile: Profile as provided by WorkOS
    """
    if not attributes:
        return

    for field_name, value in attributes.items():
        if field_name.startswith(("!", "~")) is False:
            continue
        required = field_name.startswith("!")
        field_name = field_name[1:]
        # This should not error out since "!/~" can only be used for fields that exist and have related lookups.
        field = obj._meta.get_field(field_name)
        try:
            related_instance = field.related_model.objects.get(**render_attribute(value, profile))
        except ObjectDoesNotExist:
            related_instance = None

        if required and related_instance is None:
            raise RelationDoesNotExist()

        setattr(obj, field_name, related_instance)

def _update_attributes(obj: models.Model, attributes: Optional[dict], profile: dict, template_only: bool=False) -> None:
    """
    Update an object passed in with any attributes defined. The attributes can be a template to access profile data.
    :param obj: The object to update
    :param attributes: A dictionary of attributes to update. Ex. {"username": "{{profile.first_name}}{{profile.last_name}}"}
    :param profile: The WorkOS provided profile https://workos.com/docs/reference/sso/profile
    :param template_only: If True, only update fields that rely on template and do not update static attributes.
    :return None:
    """
    if not attributes:
        return

    for field_name, value in attributes.items():
        if field_name.startswith(("!", "~")):
            continue
        try:
            field = obj._meta.get_field(field_name)
        except FieldDoesNotExist:
            # Even though this isn't a proper field it might be some custom attribute or settable property.
            # Still set the attribute if it exists on the object.
            if hasattr(obj, field_name) and (template_only is False or value != render_attribute(value, profile)):
                setattr(obj, field_name, render_attribute(value, profile))
            continue

        if field.is_relation is False:
            # If template only, do not update if the value equals the rendered value which implies it is not a template field.
            if template_only is False or value != render_attribute(value, profile):
                setattr(obj, field_name, render_attribute(value, profile))

        elif field.many_to_one or field.one_to_one:
            if value is None:
                if template_only is False:
                    setattr(obj, field_name, None)
            else:
                # Might need to create object
                related_obj = getattr(obj, field_name)
                if related_obj is None:
                    # Need to create object with the top level attributes
                    related_obj_attrs = {k: v for k,v in value.items() if not isinstance(v, dict)}
                    related_obj = field.related_model.objects.create(**related_obj_attrs)
                    setattr(obj, field_name, related_obj)

                _update_attributes(getattr(obj, field_name), value, profile)

    obj.save()


def jit_create_user(rule: models.Model, profile: dict) -> models.Model:
    """
    Given a profile from workos and a rule create a user
    """
    # Make sure if user (or related object fails) transaction rolls back
    with transaction.atomic():
        username = rule.format_username(profile)
        user = get_user_model().objects.create_user(username=username, email=profile["email"],
                                                    first_name=profile["first_name"], last_name=profile["last_name"],
                                                    is_active=False)
        user.set_unusable_password()
        _link_attributes(user, rule.saved_attributes, profile)
        _update_attributes(user, rule.saved_attributes, profile)
        # Only mark them active once we know the user has been created successfully unless explicitly set in rule.
        if "is_active" not in rule.saved_attributes:
            user.is_active = True
        user.save()

        for group in rule.jit_groups.all():
            user.groups.add(group)

    return user


def update_user_profile(user: models.Model, rule: Optional[models.Model], profile: dict) -> None:
    """
    Called on SSO login to update user attributes if anything changed on the SSO side
    :param user: user that is logging in
    :param profile: profile returned from workos that contains first, last, email
    :param rule: the rule used for the login. This should be a form of SSO since only SSO logins have attributes that
                 can update user attributes.
    """
    if conf.WORKOS_AUTO_UPDATE is True and rule:
        user.first_name = profile["first_name"]
        user.last_name = profile["last_name"]
        user.email = profile["email"]
        _update_attributes(user, rule.saved_attributes, profile, template_only=True)


def pack_state(state_dict: dict) -> str:
    """
    Pack state object into a string that will be used during handshake.
    State must be unpacked with unpack_state.
    This will sign the state so secrets cannot be seen and it will merge in
    any user defined state variables (which will not be signed).
    """
    state = signing.dumps(state_dict, compress=True)
    return json.dumps({"_": state, **conf.WORKOS_EXTRA_STATE})


def unpack_state(state_str: str) -> dict:
    """
    Unpack state that was packed with pack_state.
    Drop any extra state and only return what was originally packed.
    """
    state = json.loads(state_str)
    return signing.loads(state["_"])



# Generated by Django 4.0.6 on 2022-08-03 22:42

from django.db import migrations

DEFAULT_NAME = "Username/password login"


def create_initial_rule(apps, schema_editor):
    LoginRule = apps.get_model('workos_login', 'LoginRule')
    LoginRule.objects.create(
        name=DEFAULT_NAME,
        method="username",
        priority=1000,  # Put as a rule that is later than the default priority
        email_regex=".*"  # Match all
    )



class Migration(migrations.Migration):

    dependencies = [
        ('workos_login', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(create_initial_rule),
    ]



from django.contrib.auth.decorators import login_required
from django.urls import path
from django.views.generic import TemplateView

from .forms import WorkosPasswordResetForm, WorkosSetPasswordForm
from .views import WorkosLoginView, get_login_method, MFAVerificationView, MFAEnrollSMSView, \
    MFAEnrollTOTPView, MagicCallbackView, SSOCallbackView, MFAStartEnrollView
from django.contrib.auth import views as auth_views

urlpatterns = [
    path('verify/', MFAVerificationView.as_view(), name="mfa_verify"),
    path('enroll/', MFAStartEnrollView.as_view(), name="mfa_enroll"),
    path('enroll-sms/', MFAEnrollSMSView.as_view(), name="mfa_enroll_sms"),
    path('enroll-totp/', MFAEnrollTOTPView.as_view(), name="mfa_enroll_totp"),
    path('complete/', TemplateView.as_view(template_name="registration/magic_link_complete.html"),
         name="magic_link_confirmation"),
    path('login-method/', get_login_method, name="login_method"),
    path('callback/sso/', SSOCallbackView.as_view(), name="sso_callback"),
    path('callback/magic/', MagicCallbackView.as_view(), name="magic_callback"),
    path("login/", WorkosLoginView.as_view(), name="login"),
    path("logout/", auth_views.LogoutView.as_view(), name="logout"),
    path(
        "password_change/", auth_views.PasswordChangeView.as_view(), name="password_change"
    ),
    path(
        "password_change/done/",
        auth_views.PasswordChangeDoneView.as_view(),
        name="password_change_done",
    ),
    path("password_reset/", auth_views.PasswordResetView.as_view(
        form_class=WorkosPasswordResetForm
    ), name="password_reset"),
    path(
        "password_reset/done/",
        auth_views.PasswordResetDoneView.as_view(),
        name="password_reset_done",
    ),
    path(
        "reset/<uidb64>/<token>/",
        auth_views.PasswordResetConfirmView.as_view(form_class=WorkosSetPasswordForm),
        name="password_reset_confirm",
    ),
    path(
        "reset/done/",
        auth_views.PasswordResetCompleteView.as_view(),
        name="password_reset_complete",
    ),
]



import django.dispatch

# for jit creation workos_user_created will be called after the user is saved.
workos_user_created = django.dispatch.Signal()
workos_send_magic_link = django.dispatch.Signal()



from django.contrib.auth import REDIRECT_FIELD_NAME
from django.shortcuts import redirect
from django.urls import reverse
from django.contrib import admin, messages
from django.contrib.admin.apps import AdminConfig
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.utils.translation import gettext_lazy as _


class AdminSharedLogin(admin.AdminSite):

    @method_decorator(never_cache)
    def login(self, request, extra_context=None):
        if request.method == "GET" and self.has_permission(request):
            # Already logged-in, redirect to admin index
            index_path = reverse("admin:index", current_app=self.name)
            return redirect(index_path)

        if request.user.is_authenticated:
            # User is authenticated but does not have access to admin
            messages.warning(request, _("You are trying to access a portion of the site you do not have access to."))

        if (
                REDIRECT_FIELD_NAME not in request.GET
                and REDIRECT_FIELD_NAME not in request.POST
        ):
            next_url = reverse("admin:index", current_app=self.name)
        else:
            next_url = request.GET.get(REDIRECT_FIELD_NAME, request.POST.get(REDIRECT_FIELD_NAME))

        from django.contrib.auth.views import redirect_to_login
        return redirect_to_login(next_url)


class AdminLoginConfig(AdminConfig):
    default_site = 'workos_login.admin_site.AdminSharedLogin'



# Generated by Django 4.1.4 on 2023-02-27 00:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('workos_login', '0007_auto_20230227_0020'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='loginrule',
            name='jit_creation',
        ),
        migrations.AlterField(
            model_name='loginrule',
            name='jit_creation_type',
            field=models.CharField(blank=True, choices=[('attributes', 'Matching Attributes'), ('idp', 'Matching Attributes or IdP login')], help_text='If enabled a user account will be automatically created if one does not exist.', max_length=10, verbose_name='Just in time account creation type'),
        ),
    ]



# Generated by Django 4.0.6 on 2022-10-04 22:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('workos_login', '0003_loginrule_jit_username_format'),
    ]

    operations = [
        migrations.RenameField(
            model_name='loginrule',
            old_name='jit_attributes',
            new_name='saved_attributes',
        ),
        migrations.AlterField(
            model_name='loginrule',
            name='saved_attributes',
            field=models.JSONField(blank=True, default=dict,
                                   help_text="Attributes to set on user instance when creating user. This is a Django template string that supports lookups like {{profile.raw_attributes['some-extended-attribute']}}. <a target='_blank' href='https://workos.com/docs/reference/sso/profile'>Profile</a> is defined by WorkOS"),
        ),
    ]



from typing import Optional

from django.contrib import messages
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.auth.views import LoginView
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError
from django.http import JsonResponse, Http404, HttpRequest
from django.shortcuts import redirect, resolve_url
from django.conf import settings
from django.urls import reverse
from django.views.decorators.http import require_GET
from django.contrib.auth import login as auth_login
from django.utils.translation import gettext_lazy as _

import workos

# Create your views here.
from django.views.generic import FormView, RedirectView, TemplateView

from workos_login.forms import LoginForm, MFAVerificationForm, MFAEnrollFormSMS, MFAEnrollFormTOTP
from workos_login.models import LoginRule, UserLogin, LoginMethods
from workos_login.conf import conf
from workos_login.signals import workos_user_created, workos_send_magic_link
from workos_login.utils import user_has_mfa_enabled, get_user_login_model, mfa_enroll, jit_create_user, \
    pack_state, unpack_state, update_user_profile, find_user_by_email, get_users, has_user_login_model, find_user

SESSION_AUTHENTICATED_USER_ID = "workos_auth_user_id"  # Stores the authenticated user id (used by MFA)
SESSION_USER_ID = "workos_user_id"  # Store the user ID in SSO state.
STATE_USERNAME = "workos_username"  # Username entered - used to verify the SSO is the same name.
SESSION_RULE_ID = "workos_rule_id"  # The LoginRule ID used

# Store the safe next URL. Safeness is checked when inserting into session (or signed state object)
SESSION_NEXT = "workos_next"
SESSION_MFA_FACTOR_ID = "workos_mfa_factor_id"  # Used during MFA enrollment to store the factor ID
SESSION_TOTP_QR_CODE = "workos_totp_qr_code"  # QR Code used for enrollment
SESSION_TOTP_SECRET = "workos_totp_secret"  # Secret that matches the above QR code


def clear_session_vars(request: HttpRequest) -> None:
    if SESSION_USER_ID in request.session:
        del request.session[SESSION_USER_ID]

    if SESSION_AUTHENTICATED_USER_ID in request.session:
        del request.session[SESSION_AUTHENTICATED_USER_ID]

    if SESSION_RULE_ID in request.session:
        del request.session[SESSION_RULE_ID]

    if SESSION_NEXT in request.session:
        del request.session[SESSION_NEXT]

    if SESSION_MFA_FACTOR_ID in request.session:
        del request.session[SESSION_MFA_FACTOR_ID]

    if SESSION_TOTP_QR_CODE in request.session:
        del request.session[SESSION_TOTP_QR_CODE]

    if SESSION_TOTP_SECRET in request.session:
        del request.session[SESSION_TOTP_SECRET]


class UserNotFound(Exception):
    pass


def get_session_user(request: HttpRequest):
    user_id = request.session[SESSION_AUTHENTICATED_USER_ID]
    if(user_id):
        return get_users().get(pk=user_id)
    return None


def get_session_rule(request: HttpRequest) -> Optional[LoginRule]:
    rule_id = request.session[SESSION_RULE_ID]
    if rule_id:
        return LoginRule.objects.get(pk=rule_id)
    return None


def login_session_user(request: HttpRequest):
    user = get_session_user(request)

    if not user:
        raise UserNotFound

    auth_login(request, user)
    clear_session_vars(request)
    return user


@require_GET
def get_login_method(request: HttpRequest):
    username = request.GET.get("username", default="").strip()
    rule_id = None
    login_method = LoginMethods.USERNAME
    if not username:
        return JsonResponse({"message": "Username is required"}, status=400)

    rule = LoginRule.objects.find_rule_for_username(username)
    if rule:
        method = LoginMethods(rule.method)
        login_method = method.frontend_method
        if not method.is_sso:
            # If method is not SSO check if MFA is enabled, if so that is the method we will use.
            # Since the user has enabled MFA for themselves.
            try:
                user = find_user(username)
                if user_has_mfa_enabled(user):
                    login_method = LoginMethods.MFA
            except ObjectDoesNotExist:
                pass


    return JsonResponse({"method": login_method}, status=200)


class BaseCallbackView(RedirectView):
    query_string = False
    permanent = False
    allow_idp_initiated = False

    def create_error(self, msg: str) -> str:
        messages.error(self.request, msg)
        return reverse("login")

    def update_user_login(self, user_login: UserLogin, workos_profile: dict) -> None:
        """
        Can be used to update the user login profile
        :param user_login: the UserLogin object for this user
        :param workos_profile: the workos profile
        """
        # Save the workos id
        user_login.sso_id = workos_profile["id"]
        user_login.save()

    def find_user_login(self, workos_profile: dict) -> Optional[UserLogin]:
        try:
            user_login = UserLogin.objects.get(sso_id=workos_profile["id"])
        except UserLogin.DoesNotExist:
            user_login = None
        return user_login

    def get_redirect_url(self, *args, **kwargs):
        code = self.request.GET.get("code")
        state = self.request.GET.get("state")
        user = None
        username = None
        if state:
            state = unpack_state(state)

        profile = workos.client.sso.get_profile_and_token(code).to_dict()["profile"]

        user_login = self.find_user_login(profile)

        # Default to base redirect login
        next_url = settings.LOGIN_REDIRECT_URL

        if user_login:
            # User has logged in before, and we know the user account
            user = user_login.user
            rule = user_login.rule

            if state:
                next_url = state[SESSION_NEXT]
                if user.pk != state[SESSION_USER_ID]:
                    return self.create_error(_("Your username does not match the account you selected."))
        elif state:
            # Login based on a rule, but this is the first time a user is loging in
            # via SSO (since there is no user_login yet) or it is JIT creation.
            rule_id = state[SESSION_RULE_ID]
            user_id = state[SESSION_USER_ID]
            next_url = state[SESSION_NEXT]
            username = state[STATE_USERNAME]
            if user_id:
                user = get_users().get(pk=user_id)

                if has_user_login_model(user):
                    return self.create_error(_("Selected account does not match the user account"))
                elif not user.email or user.email.lower() != profile["email"].lower():
                    # We have a user on file and this is the first time they are logging in via SSO (since there is no user login).
                    # But the email does not match. For the first association the email address must match
                    # to ensure the user is logging into the account they are supposed to.
                    return self.create_error(_("Your email address does not match your user account on file. Please try again"))

            rule = LoginRule.objects.get(pk=rule_id)
        elif self.allow_idp_initiated:
            # First time IDP Initiated
            # If they have logged in before we would have a user_login created already
            # Since this is a first time login we have not created a formatted username yet.
            # Instead, do a lookup based on email to try to find a matching account already in the system.
            user = find_user_by_email(profile["email"])
            if user:
                rule = LoginRule.objects.find_rule_for_user(user)
            else:
                # No user found, but maybe a JIT rule exists for this email address
                rule = LoginRule.objects.find_rule_for_jit(profile["email"],
                                                           connection_id=profile.get("connection_id"),
                                                           organization_id=profile.get("organization_id"))
            if rule is None:
                return self.create_error(_("No login was found, please contact your administrator"))
            if rule.organization_id != profile["organization_id"] and rule.connection_id != profile["connection_id"]:
                return self.create_error(_("Unknown SSO connection or organization"))
        else:
            return self.create_error(_("Unable to login. Please contact your administrator."))

        if user is None:
            if not rule.jit_creation_type:
                return self.create_error(_("Your user account has not been provisioned. Please contact your administrator to create an account."))

            # Since there is no user this must have been an email match, make sure
            # the email they used is for the correct rule (ensuring they didn't select a wrong account)
            expected_rule = LoginRule.objects.find_rule_for_jit(profile["email"],
                                                                connection_id=profile.get("connection_id"),
                                                                organization_id=profile.get("organization_id"))

            if rule != expected_rule:
                return self.create_error(_("Account username does not match your request. Please try again"))

            # If there is a username but no user it is because we are doing an SP JIT login
            # that matched an email regex rule. Make sure that the email associated with the
            # SSO account matches the username the user entered (which would be their email).
            if username and username != profile["email"]:
                return self.create_error(_("The account selected does not match username entered. Please try again"))

            try:
                user = jit_create_user(rule, profile)
            except IntegrityError:
                return self.create_error(_("Unable to create user account, you may already have an account created."))
            workos_user_created.send(sender=UserLogin, user=user, profile=profile, rule=rule)
        else:
            try:
                update_user_profile(user, rule, profile)
            except IntegrityError:
                self.create_error(_("Unable to update user profile, please contact your administrator."))

        # We have a user, make sure they are active
        if not user.is_active:
            return self.create_error(_("Your user account is not active. Please contact your administrator."))

        if not user_login:
            # First time through get the user login and set the rule that is being used.
            user_login = get_user_login_model(user)
            user_login.rule = rule
            user_login.save()

        self.update_user_login(user_login, profile)
        # Login user and save sso attributes
        # Check if the user is active and has permissions
        auth_login(self.request, user)
        clear_session_vars(self.request)

        return next_url


class MagicCallbackView(BaseCallbackView):
    pass


class SSOCallbackView(BaseCallbackView):
    allow_idp_initiated = True

    def find_user_login(self, workos_profile: dict) -> Optional[UserLogin]:
        try:
            user_login = UserLogin.objects.get(sso_id=workos_profile["id"])
        except UserLogin.DoesNotExist:
            user_login = None
        return user_login

    def update_user_login(self, user_login: UserLogin, workos_profile: dict) -> None:
        user_login.mfa_type = ""  # Don't have MFA enabled if using SSO - rely on SSO to do this
        user_login.sso_id = workos_profile["id"]
        user_login.idp_id = workos_profile["idp_id"]
        user_login.save()


class WorkosLoginView(LoginView):
    form_class = LoginForm

    def __init__(self, *args, **kwargs):
        self.cookies_disabled = False
        super().__init__(*args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["login_home"] = resolve_url(settings.LOGIN_REDIRECT_URL)
        return ctx

    def get(self, request, *args, **kwargs):
        # Starting a new login - clear session
        clear_session_vars(request)
        request.session.set_test_cookie()
        return super().get(request, *args, **kwargs)

    def form_valid(self, form):
        # First check if cookie are enabled
        if not self.request.session.test_cookie_worked():
            messages.error(self.request, _("Cookies are required for proper functionality. Please enable cookies and try again."))
            self.cookies_disabled = True
            return self.form_invalid(form)

        user = form.get_user()
        rule = form.get_rule()
        method = rule.method
        state = pack_state({
            SESSION_RULE_ID: rule.pk,
            SESSION_USER_ID: user.pk if user else 0,
            STATE_USERNAME: form.cleaned_data["username"],
            SESSION_NEXT: self.get_success_url()
        })

        self.request.session[SESSION_RULE_ID] = rule.pk
        self.request.session[SESSION_NEXT] = self.get_success_url()
        # Check MFA before email to see if the user enabled MFA even without
        # having it as part of a rule
        if method == LoginMethods.MFA or user_has_mfa_enabled(user):
            self.request.session[SESSION_AUTHENTICATED_USER_ID] = user.pk
            if user_has_mfa_enabled(user):
                return redirect('mfa_verify')
            else:
                return redirect('mfa_enroll')
        elif method == LoginMethods.USERNAME:
            # This will log in the user - clear the workos session vars
            clear_session_vars(self.request)
            return super(WorkosLoginView, self).form_valid(form)
        elif method == LoginMethods.MAGIC_LINK or method == LoginMethods.EMAIL_MFA:
            email = user.email
            uri = conf.WORKOS_MAGIC_REDIRECT_URI if conf.WORKOS_MAGIC_REDIRECT_URI else self.request.build_absolute_uri(reverse("magic_callback"))
            session = workos.client.passwordless.create_session({
                'email': email,
                'redirect_uri': uri,
                'state': state,
                'type': 'MagicLink'})
            workos_send_magic_link.send(sender=LoginRule, user=user, link=session["link"], rule=rule)
            if not conf.WORKOS_SEND_CUSTOM_EMAIL:
                workos.client.passwordless.send_session(session['id'])
            return redirect('magic_link_confirmation')
        elif method == LoginMethods.SAML_SSO:
            uri = conf.WORKOS_SSO_REDIRECT_URI if conf.WORKOS_SSO_REDIRECT_URI else self.request.build_absolute_uri(reverse("sso_callback"))
            authorization_url = workos.client.sso.get_authorization_url(
                connection=rule.connection_id if rule.connection_id else None,
                organization=rule.organization_id if rule.organization_id else None,
                redirect_uri=uri,
                state=state,
            )
            return redirect(authorization_url)
        elif method in {LoginMethods.MICROSOFT_SSO, LoginMethods.GOOGLE_SSO}:
            uri = conf.WORKOS_SSO_REDIRECT_URI if conf.WORKOS_SSO_REDIRECT_URI else self.request.build_absolute_uri(reverse("sso_callback"))
            authorization_url = workos.client.sso.get_authorization_url(
                redirect_uri=uri,
                provider=LoginMethods(method).provider,
                state=state,
            )
            return redirect(authorization_url)

        # Unknown method
        raise Http404


class MFAPermissionMixin(UserPassesTestMixin):
    def test_func(self):
        if self.request.user.is_authenticated:
            return True

        if self.request.session.get(SESSION_AUTHENTICATED_USER_ID, False):
            return True

        return False


class LoginSuccessUrlMixin:
    def get_success_url(self):
        if SESSION_NEXT in self.request.session:
            return self.request.session[SESSION_NEXT]

        return settings.LOGIN_REDIRECT_URL


class MFAVerificationView(MFAPermissionMixin, LoginSuccessUrlMixin, FormView):
    form_class = MFAVerificationForm
    template_name = 'registration/mfa_verify.html'

    def get_factor_id(self):

        if SESSION_MFA_FACTOR_ID in self.request.session:
            return self.request.session[SESSION_MFA_FACTOR_ID]

        user_login = get_user_login_model(get_session_user(self.request))
        if not user_login:
            raise Http404

        factor_id = user_login.mfa_factor
        if not factor_id:
            # No factor that we can authenticate
            raise Http404

        return factor_id

    def __init__(self, *args, **kwargs):
        self.workos_response = None
        super(MFAVerificationView, self).__init__(*args, **kwargs)

    def get(self, request, *args, **kwargs):
        # Only on first GET do we get a challenge factor - if posting it will already be in the form data
        message = conf.WORKOS_SMS_MFA_TEMPLATE
        if callable(message):
            message = message(get_session_user(self.request))
        self.workos_response = workos.client.mfa.challenge_factor(authentication_factor_id=self.get_factor_id(), sms_template=message)
        return super(MFAVerificationView, self).get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super(MFAVerificationView, self).get_context_data(**kwargs)
        user_login = get_user_login_model(get_session_user(self.request))
        ctx["sms"] = user_login.is_sms
        ctx["totp"] = user_login.is_totp
        return ctx

    def get_initial(self):
        # Add the challenge ID if we don't have it yet
        initial = super(MFAVerificationView, self).get_initial()
        if self.workos_response:
            initial["challenge_id"] = self.workos_response["id"]
        return initial

    def form_valid(self, form):
        if not self.request.user.is_authenticated:
            # They could be enrolling/verifying while logged in - only log in if needed
            login_session_user(self.request)

        if SESSION_MFA_FACTOR_ID in self.request.session:
            # We are enrolling so we want to save the factor ID by sending signal to app
            # Update factor id
            mfa_enroll(self.request.user, self.request.session[SESSION_MFA_FACTOR_ID], mfa_type="sms")
            del self.request.session[SESSION_MFA_FACTOR_ID]
            # user is already signed in for this case - no need for login

        return super(MFAVerificationView, self).form_valid(form)


class MFAStartEnrollView(MFAPermissionMixin, TemplateView):
    template_name = "registration/mfa_enroll_start.html"


class MFAEnrollBaseView(MFAPermissionMixin, FormView):
    template_name = 'registration/mfa_enroll.html'

    def get_success_url(self):
        return reverse('mfa_verify')

    def form_valid(self, form):
        self.request.session[SESSION_MFA_FACTOR_ID] = form.complete_enrollment(self.request.user)
        return super(MFAEnrollBaseView, self).form_valid(form)


class MFAEnrollSMSView(MFAEnrollBaseView):
    form_class = MFAEnrollFormSMS

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["is_sms"] = True
        return ctx


class MFAEnrollTOTPView(LoginSuccessUrlMixin, FormView):
    template_name = 'registration/mfa_enroll.html'
    form_class = MFAEnrollFormTOTP

    def __init__(self, *args, **kwargs):
        super(MFAEnrollTOTPView, self).__init__(*args, **kwargs)

    def get(self, request, *args, **kwargs):
        user = get_session_user(request)
        rule = get_session_rule(request)
        if not user or not rule:
            messages.error(request, _("Unauthorized - try again"))
            return redirect("login")
        if SESSION_TOTP_QR_CODE not in self.request.session:
            workos_response = workos.client.mfa.enroll_factor(type=conf.MFA_TOTP_TYPE,
                                                                   totp_issuer=rule.totp_organization_name,
                                                                   totp_user=user.email)
            self.request.session[SESSION_TOTP_QR_CODE] = workos_response["totp"]["qr_code"]
            self.request.session[SESSION_TOTP_SECRET] = workos_response["totp"]["secret"]
            self.request.session[SESSION_MFA_FACTOR_ID] = workos_response["id"]
        return super(MFAEnrollTOTPView, self).get(request, *args, **kwargs)

    def form_valid(self, form):
        mfa_enroll(get_session_user(self.request), form.cleaned_data["factor_id"], mfa_type="totp")
        login_session_user(self.request)
        return super(MFAEnrollTOTPView, self).form_valid(form)

    def get_initial(self):
        initial = super(MFAEnrollTOTPView, self).get_initial()
        initial["factor_id"] = self.request.session[SESSION_MFA_FACTOR_ID]
        return initial

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["is_totp"] = True
        return ctx

# Allow for an API post (open to all) to post a username/email to get the flow started.



# Generated by Django 4.0.6 on 2022-10-05 06:11

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('workos_login', '0004_rename_jit_attributes_loginrule_saved_attributes'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='loginrule',
            name='jit_username_format',
        ),
        migrations.AddField(
            model_name='userlogin',
            name='rule',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='workos_login.loginrule'),
        ),
    ]



from enum import Enum
from typing import Optional

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django import forms
from django.utils.translation import gettext_lazy as _
from workos.utils.connection_types import ConnectionType

from workos_login.conf import conf
import re

# Create your models here.
from workos_login.exceptions import UnknownUsernameFormat
from workos_login.utils import find_user, render_attribute, get_users


class EmailRegexField(models.TextField):

    def formfield(self, **kwargs):
        return super().formfield(widget=forms.TextInput)


class WorkosQuerySet(models.QuerySet):
    def find_rule_for_username(self, username_or_email: str) -> Optional["LoginRule"]:
        """Find a rule for a matching username"""
        for rule in self.all():
            if rule.rule_applies_to_username(username_or_email):
                return rule
        return None

    def find_rule_for_user(self, user: models.Model) -> Optional["LoginRule"]:
        """
        Find the applied rule for this user
        :param user: The user to test
        :return: The rule that applies - it will only return 1 rule at most.
        """
        for rule in self.all():
            if rule.rule_applies_to_user(user):
                return rule
        return None

    def find_rule_for_jit(self, email: str, connection_id: Optional[str], organization_id: Optional[str]) -> Optional["LoginRule"]:
        """Find a rule for a matching an email address
        This could return a rule that matches an email address even if JIT is not enabled and the user does not exist.
        This is used for IdP initiated logins to try and find a matching rule before associated to user or
        creating user (if JIT is enabled).
        """
        for rule in self.all():
            connection_match = (rule.organization_id == organization_id or rule.connection_id == connection_id)
            if rule.jit_creation_type == JitMethods.IDP and connection_match:
                return rule
            # For ATTRIBUTES_MATCH connection_match is not checked since it could be google or microsoft login.
            if rule.jit_creation_type == JitMethods.ATTRIBUTES_MATCH and rule.test_email(email):
                return rule

        return None


class LoginMethods(models.TextChoices):
    MFA = conf.WORKOS_METHOD_MFA, "MFA"
    EMAIL_MFA = conf.WORKOS_METHOD_EMAIL_MFA, "Email MFA"
    MAGIC_LINK = conf.WORKOS_METHOD_MAGIC, "Passwordless Email"
    GOOGLE_SSO = conf.WORKOS_METHOD_GOOGLE_OAUTH, "Google SSO"
    MICROSOFT_SSO = conf.WORKOS_METHOD_MICROSOFT_OAUTH, "Microsoft SSO"
    SAML_SSO = conf.WORKOS_METHOD_SSO, "SAML SSO"
    USERNAME = conf.WORKOS_METHOD_EMAIL, "Username/Password"

    @property
    def is_sso(self):
        return self == self.GOOGLE_SSO or self == self.MICROSOFT_SSO or self == self.SAML_SSO

    @property
    def frontend_method(self):
        return "sso" if self.is_sso else self

    @property
    def needs_con_or_org_id(self):
        return self == self.SAML_SSO

    @property
    def provider(self):
        if self == self.GOOGLE_SSO:
            return ConnectionType.GoogleOAuth
        if self == self.MICROSOFT_SSO:
            return ConnectionType.MicrosoftOAuth
        return None


class JitMethods(models.TextChoices):
    ATTRIBUTES_MATCH = "attributes", "Matching Attributes"
    IDP = "idp", "Matching Attributes or IdP login"

class LoginRule(models.Model):
    name = models.CharField(max_length=255, help_text=_("Name for this config"), unique=True)
    lookup_attributes = models.JSONField(blank=True, null=True)
    totp_organization_name = models.CharField(max_length=255, blank=True, help_text=_("The name of organization which "
                                                                                    "shows in authenticator apps."))
    method = models.CharField(choices=LoginMethods.choices, max_length=15)

    jit_creation_type = models.CharField(choices=JitMethods.choices, verbose_name=_("Just in time account creation type"),
                                         max_length=10, blank=True, help_text=_("If enabled a user account will be automatically created "
                                                   "if one does not exist."))
    priority = models.IntegerField(default=100, help_text=_("Priority of this rule. Lower numbers are checked first"),
                                   unique=True)
    email_regex = EmailRegexField(blank=True, help_text=_(r"Regex to test on email. Ex. ^.+@(domain|domain2)\.com$"),
                                  verbose_name="Email Regular Expression Pattern")
    connection_id = models.CharField(max_length=255, blank=True, help_text=_("WorkOS connection ID"))
    organization_id = models.CharField(max_length=255, blank=True, help_text=_("WorkOS organization ID"))

    # Need a way to set attributes like is_external to True when setting up JIT
    jit_groups = models.ManyToManyField("auth.Group", blank=True,
                                        help_text=_("Groups a user should be added to if user is being created."))

    saved_attributes = models.JSONField(blank=True, default=dict,
                                        help_text=_("Attributes to set on user instance when creating user. This is a Django template string that supports lookups like "
                                                    "{{profile.raw_attributes['some-extended-attribute']}}. <a target='_blank' href='https://workos.com/docs/reference/sso/profile'>Profile</a> is defined by WorkOS"))
    objects = WorkosQuerySet.as_manager()

    class Meta:
        ordering = ["priority"]

    def __str__(self):
        return self.name

    @property
    def sso(self) -> bool:
        return LoginMethods(self.method).is_sso

    @property
    def mfa(self) -> bool:
        return self.method == LoginMethods.MFA

    @property
    def magic_link(self) -> bool:
        return self.method == LoginMethods.MAGIC_LINK

    @property
    def username(self) -> bool:
        return self.method == LoginMethods.USERNAME

    @property
    def requires_password(self) -> bool:
        return self.mfa or self.username or self.method == LoginMethods.EMAIL_MFA

    def clean(self):
        errors = {}

        if self.totp_organization_name and not self.mfa:
            errors["totp_organization_name"] = _("Organization name only valid for MFA")

        if self.mfa and not self.totp_organization_name:
            errors["totp_organization_name"] = _("Must set organization name if using MFA. This will appear on "
                                                 "authenticator apps")

        if self.organization_id and not self.sso:
            errors["organization_id"] = _("Organization ID is only used for SSO")

        if self.connection_id and not self.sso:
            errors["connection_id"] = _("Connection ID is only used for SSO")

        if self.connection_id and self.organization_id:
            raise ValidationError(_("You cannot set both an organization ID and connection ID"))

        if LoginMethods(self.method).needs_con_or_org_id and not (self.organization_id or self.connection_id):
            errors["connection_id"] = _("For SSO you must provide either connection ID or organization ID")
            errors["organization_id"] = _("For SSO you must provide either connection ID or organization ID")

        if not self.jit_creation_type and not self.lookup_attributes and not self.email_regex:
            errors["lookup_attributes"] = errors["email_regex"] = _("This rule does not apply to any users. "
                                                                    "Set lookup attributes or an email pattern.")

        if self.jit_creation_type and not self.sso:
            errors["jit_creation_type"] = _("You can only set account creation if using SSO")


        if self.saved_attributes and not self.sso:
            errors["saved_attributes"] = _("Custom attributes only apply to SSO login.")

        if errors:
            raise ValidationError(errors)

    def test_email(self, email: str) -> bool:
        if self.email_regex and re.search(self.email_regex, email, flags=re.IGNORECASE):
            return True
        return False

    def rule_applies_to_username(self, username_or_email: str) -> bool:
        """
        Given a username try to find the rule that applies
        :param username_or_email: A username or email to lookup
        :return: True if the rule applies False otherwise
        """
        # Check if a user exists
        user = find_user(username_or_email)
        if user:
            return self.rule_applies_to_user(user)

        if not self.jit_creation_type:
            return False

        if self.test_email(username_or_email):
            return True

        return False

    def rule_applies_to_user(self, user: models.Model):
        exists = False
        if self.lookup_attributes:
            # Check two special keys first "has_mfa" checks
            attrs = self.lookup_attributes
            has_mfa = attrs.pop("has_mfa", False)
            if has_mfa:
                exists = UserLogin.objects.filter(user=user).exclude(mfa_factor="").exists()

            has_sso = attrs.pop("has_sso", False)
            if has_sso and not exists:
                exists = UserLogin.objects.filter(user=user).exclude(sso_id="").exists()

            if not exists and attrs:
                exists = get_users().filter(**attrs).filter(pk=user.pk).exists()

        if not exists:
            # The user attributes do not match, check if email domain matches
            return self.test_email(user.email)

        return exists

    def format_username(self, profile: dict) -> str:
        if "username" in self.saved_attributes:
            return render_attribute(self.saved_attributes["username"], profile)

        if conf.WORKOS_JIT_USERNAME == "email":
            return profile["email"]
        if conf.WORKOS_JIT_USERNAME == "idp_id":
            return profile["idp_id"]
        if conf.WORKOS_JIT_USERNAME == "id":
            return profile["id"]
        raise UnknownUsernameFormat

class UserLogin(models.Model):
    """One-to-one link to a user that sets the ids needed for sso and mfa"""
    mfa_factor = models.CharField(max_length=255, blank=True)
    MFA_TYPE_CHOICES = (
        (conf.MFA_SMS_TYPE, "SMS"),
        (conf.MFA_TOTP_TYPE, "Authenticator App")
    )
    mfa_type = models.CharField(choices=MFA_TYPE_CHOICES, max_length=10, blank=True)
    sso_id = models.CharField(max_length=255, blank=True)
    idp_id = models.CharField(max_length=255, blank=True)
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    # Rule associated with UserLogin - used for SSO particularly for being able to update saved_attributes
    rule = models.ForeignKey(LoginRule, blank=True, null=True, on_delete=models.CASCADE)
    created_at = models.DateTimeField(auto_now_add=True)
    last_modified = models.DateTimeField(auto_now=True)

    def __str__(self):
        return str(self.user)

    @property
    def mfa_enabled(self):
        return bool(self.mfa_type) and bool(self.mfa_factor)

    @property
    def is_sms(self):
        return self.mfa_type == conf.MFA_SMS_TYPE

    @property
    def is_totp(self):
        return self.mfa_type == conf.MFA_TOTP_TYPE




from django.contrib.auth.forms import AuthenticationForm, UsernameField, PasswordResetForm, SetPasswordForm
from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _
from workos.exceptions import BadRequestException, ServerException

from workos_login.conf import conf
from django.core.exceptions import ValidationError
import workos

from workos_login.models import LoginRule
from workos_login.utils import find_user, verify_challenge, totp_verify_code, user_has_mfa_enabled


class BootstrapMixin:

    def __init__(self, *args, **kwargs):
        kwargs["label_suffix"] = ""
        super().__init__(*args, **kwargs)
        for name, field in self.fields.items():
            field.widget.attrs["class"] = "form-control"

    def _post_clean(self):
        for name, field in self.fields.items():
            class_list = ["form-control"]
            if(field.get_bound_field(self, name).errors):
                class_list.append("is-invalid")
            field.widget.attrs["class"] = " ".join(class_list)


class LoginForm(BootstrapMixin, AuthenticationForm):

    username = UsernameField(label=_("Username"), widget=forms.TextInput(attrs={"autofocus": True}))
    password = forms.CharField(
        label=_("Password"),
        strip=False,
        required=False,
        widget=forms.PasswordInput(attrs={"autocomplete": "current-password"}),
    )

    def __init__(self, *args, **kwargs):
        self.rule_cache = None
        super(LoginForm, self).__init__(*args, **kwargs)
        self.fields["username"].widget.attrs["class"] = "form-control"
        if conf.WORKOS_EMAIL_LOOKUP:
            self.fields["username"].label = _("Username or Email")


    def get_rule(self):
        return self.rule_cache

    def clean(self):
        rule = LoginRule.objects.find_rule_for_username(self.cleaned_data["username"])
        if not rule:
            # No rule for user - cannot login
            # there should always be at least one rule
            raise self.get_invalid_login_error()
        self.rule_cache = rule

        try:
            user = find_user(self.cleaned_data.get("username"))
        except ObjectDoesNotExist:
            user = None

        if rule.requires_password or (user and user_has_mfa_enabled(user)):
            # Even if the rule doesn't require a password but the user has mfa enabled, that will take precedence.
            # MFA is always checked before magic email and will be respected to require MFA.
            if not self.cleaned_data.get("password"):
                raise self.get_invalid_login_error()
            return super().clean()

        self.user_cache = user

        if rule.magic_link:
            if not user:
                raise self.get_invalid_login_error()
            self.confirm_login_allowed(user)
        if rule.sso:
            # Make sure that there is either an active user or that there is a JIT
            if not rule.sso:
                raise self.get_invalid_login_error()
            if not rule.jit_creation_type:
                if not user:
                    raise self.get_invalid_login_error()
                self.confirm_login_allowed(user)

        return self.cleaned_data


class MFAVerificationForm(BootstrapMixin, forms.Form):
    """
    Used to verify an already enrolled MFA user.
    """
    code = forms.CharField(required=True, widget=forms.TextInput(attrs={"autofocus": True}))
    challenge_id = forms.CharField(widget=forms.HiddenInput(), required=True)

    def clean(self):
        challenge_id = self.cleaned_data["challenge_id"]
        challenge_code = self.cleaned_data["code"]
        if not verify_challenge(challenge_id, challenge_code):
            self.add_error("code", ValidationError(_("Verification failed - please try again"), code="invalid_code"))

        return self.cleaned_data


class MFAEnrollFormSMS(BootstrapMixin, forms.Form):
    """Used to set a new enrollment"""
    phone_number = forms.CharField(required=True, help_text="Enter the phone number you would like to use "
                                                            "to receive your text message authentication codes")

    def __init__(self, *args, **kwargs):
        self.response_id = None
        super(MFAEnrollFormSMS, self).__init__(*args, **kwargs)
        
    def clean_phone_number(self):
        try:
            response = workos.client.mfa.enroll_factor(type=conf.MFA_SMS_TYPE, phone_number=self.cleaned_data["phone_number"])
        except BadRequestException as e:
            try:
                message = e.message
            except AttributeError:
                message = "Error, please try again"
            raise ValidationError(message)
        except ServerException:
            message = _("Unknown error, please try again")
            raise ValidationError(message)
        
        self.response_id = response["id"]
        
    def complete_enrollment(self, user):
        return self.response_id


class MFAEnrollFormTOTP(BootstrapMixin, forms.Form):
    """Used to set a new enrollment"""
    factor_id = forms.CharField(required=True, widget=forms.HiddenInput())
    code = forms.CharField(required=True, widget=forms.TextInput(attrs={"autofocus": True}))

    def clean(self):
        if not totp_verify_code(self.cleaned_data["factor_id"], self.cleaned_data["code"]):
            self.add_error("code", ValidationError(_("Verification failed - please try again")))

    def complete_enrollment(self, user):
        return self.cleaned_data["factor_id"]


class WorkosPasswordResetForm(BootstrapMixin, PasswordResetForm):
    pass


class WorkosSetPasswordForm(BootstrapMixin, SetPasswordForm):
    pass



# Generated by Django 4.0.6 on 2022-09-20 04:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('workos_login', '0002_auto_20220803_2242'),
    ]

    operations = [
        migrations.AddField(
            model_name='loginrule',
            name='jit_username_format',
            field=models.CharField(blank=True, choices=[('email', 'Email Address'), ('idp', 'Identity Provider ID'), ('workos', 'Unique ID')], help_text='Username format for created accounts', max_length=10),
        ),
    ]



# Generated by Django 4.1.4 on 2023-02-27 00:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('workos_login', '0005_remove_loginrule_jit_username_format_userlogin_rule'),
    ]

    operations = [
        migrations.AddField(
            model_name='loginrule',
            name='jit_creation_type',
            field=models.CharField(blank=True, choices=[('attributes', 'Matching Attributes'), ('idp', 'Matching Attributes or IdP login')], max_length=10),
        ),
    ]



# Generated by Django 4.1.4 on 2023-02-27 00:20

from django.db import migrations


def forward(apps, schema_editor):
    LoginRule = apps.get_model('workos_login', 'LoginRule')
    LoginRule.objects.filter(jit_creation=True).update(jit_creation_type="attributes")

class Migration(migrations.Migration):

    dependencies = [
        ('workos_login', '0006_loginrule_jit_creation_type'),
    ]

    operations = [
        migrations.RunPython(forward, elidable=True),
    ]



from django.apps import AppConfig
from workos_login.conf import conf
from django.core.checks import register, Error, Tags


def check_settings(app_configs, **kwargs):
    errors = []
    if conf.WORKOS_JIT_USERNAME not in ["email", "id", "idp_id"]:
        errors.append(
            Error(
                'Invalid settings WORKOS_JIT_USERNAME unknown value: {}'.format(conf.WORKOS_JIT_USERNAME),
                hint='You must set this setting to either "email", "id" or "idp_id"',
                obj=conf,
                id='workos_login.E001',
            )
        )
    return errors

class WorkosConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'workos_login'

    def ready(self):
        import workos
        # If settings do not exist they must be set in env variables
        if(conf.WORKOS_CLIENT_ID):
            workos.client_id = conf.WORKOS_CLIENT_ID
        if(conf.WORKOS_API_KEY):
            workos.api_key = conf.WORKOS_API_KEY

        register(check_settings, Tags.compatibility)




from django.utils.html import format_html
from workos.exceptions import BadRequestException
from django.utils.translation import gettext_lazy as _

from workos_login.models import LoginRule, UserLogin

from django.contrib import admin
from workos import client as workos_client


class LoginRuleAdmin(admin.ModelAdmin):
    list_display = ("name", "priority", "method")
    filter_horizontal = ("jit_groups",)
    readonly_fields = ('portal_link',)
    fieldsets = (
        (None, {
            'fields': ('name', 'method', 'priority')
        }),
        ('Users this rule applies to', {
            'fields': ('email_regex', 'lookup_attributes',)
        }),
        ('SSO Options', {
            'fields': ('connection_id', 'organization_id', 'saved_attributes', 'jit_creation_type', 'jit_groups', 'portal_link')
        }),
        ('MFA Options', {
            'fields': ('totp_organization_name',)
        }),
    )

    @admin.display(
        description='Portal Link',
    )
    def portal_link(self, obj):
        if obj and obj.organization_id:
            try:
                portal_link = workos_client.portal.generate_link(
                    organization=obj.organization_id, intent="sso",
                )
                return format_html("<a href='{}'>Link to WorkOS Portal</a>", portal_link["link"])
            except BadRequestException:
                return _("You cannot have a portal link unless you set an Admin Portal Redirect Link")
        return _("Only available for SSO with organization ID set.")


class UserLoginAdmin(admin.ModelAdmin):
    readonly_fields = fields = ("mfa_factor", "mfa_type", "sso_id", "idp_id", "user", "rule", "created_at", "last_modified")


admin.site.register(LoginRule, LoginRuleAdmin)
admin.site.register(UserLogin, UserLoginAdmin)



from typing import Optional

from django.conf import settings


class Settings:
    """
    Shadow Django's settings with defaults
    """
    @property
    def WORKOS_CLIENT_ID(self) -> str:
        return getattr(settings, "WORKOS_CLIENT_ID", None)

    @property
    def WORKOS_API_KEY(self) -> str:
        return getattr(settings, "WORKOS_API_KEY", None)

    @property
    def WORKOS_EMAIL_LOOKUP(self) -> bool:
        """Lookup based on email address - the application must enforce lowercase unique emails if this is True"""
        return getattr(settings, "WORKOS_EMAIL_LOOKUP", False)

    @property
    def WORKOS_USERNAME_LOOKUP(self) -> bool:
        """Lookup based on username"""
        return getattr(settings, "WORKOS_USERNAME_LOOKUP", True)

    @property
    def WORKOS_METHOD_EMAIL(self) -> str:
        return "username"

    @property
    def WORKOS_METHOD_SSO(self) -> str:
        return "sso"

    @property
    def WORKOS_METHOD_GOOGLE_OAUTH(self) -> str:
        return "GoogleOAuth"

    @property
    def WORKOS_METHOD_MICROSOFT_OAUTH(self) -> str:
        return "MicrosoftOAuth"

    @property
    def WORKOS_METHOD_MFA(self) -> str:
        return "mfa"

    @property
    def WORKOS_METHOD_EMAIL_MFA(self) -> str:
        return "email_mfa"

    @property
    def WORKOS_METHOD_MAGIC(self) -> str:
        return "magic"

    @property
    def WORKOS_SSO_REDIRECT_URI(self) -> Optional[str]:
        return getattr(settings, "WORKOS_SSO_REDIRECT_URI", None)

    @property
    def WORKOS_MAGIC_REDIRECT_URI(self) -> Optional[str]:
        return getattr(settings, "WORKOS_MAGIC_REDIRECT_URI", None)

    @property
    def MFA_TOTP_TYPE(self) -> str:
        return "totp"

    @property
    def MFA_SMS_TYPE(self) -> str:
        return "sms"

    @property
    def WORKOS_SMS_MFA_TEMPLATE(self) -> str:
        return getattr(settings, "WORKOS_SMS_MFA_TEMPLATE", "Your authentication code is {{code}}")

    @property
    def WORKOS_SEND_CUSTOM_EMAIL(self) -> bool:
        return getattr(settings, 'WORKOS_SEND_CUSTOM_EMAIL', False)

    @property
    def WORKOS_EXTRA_STATE(self) -> dict:
        return getattr(settings, 'WORKOS_EXTRA_STATE', {})

    @property
    def WORKOS_AUTO_UPDATE(self) -> bool:
        return getattr(settings, 'WORKOS_AUTO_UPDATE', True)

    @property
    def WORKOS_JIT_USERNAME(self) -> bool:
        # Available methods are 'email', 'idp_id', 'id'
        return getattr(settings, 'WORKOS_JIT_USERNAME', 'email')

    @property
    def WORKOS_ACTIVE_USER_FILTER(self) -> dict:
        return getattr(settings, 'WORKOS_ACTIVE_USER_FILTER', {"is_active": True})


conf = Settings()
